/*
 * Created on 04-Sep-2005
 * Created by Paul Gardner
 * Copyright (C) 2005 Aelitis, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * AELITIS, SARL au capital de 30,000 euros
 * 8 Allee Lenotre, La Grille Royale, 78600 Le Mesnil le Roi, France.
 *
 */

package org.parg.azureus.plugins.sudoku.model.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.*;

import org.parg.azureus.plugins.sudoku.model.SudokuModel;
import org.parg.azureus.plugins.sudoku.model.SudokuModelComplexity;
import org.parg.azureus.plugins.sudoku.model.SudokuModelGenerationListener;
import org.parg.azureus.plugins.sudoku.model.SudokuModelListener;
import org.parg.azureus.plugins.sudoku.model.SudokuPossibilityMetric;
import org.parg.azureus.plugins.sudoku.model.SudokuSolveInstance;
import org.parg.azureus.plugins.sudoku.model.SudokuValue;
import org.parg.azureus.plugins.sudoku.model.SudokuValueListener;
import org.parg.azureus.plugins.sudoku.model.SudokuValueMetric;

public class 
SudokuModelImpl
	implements SudokuModel, SudokuValueListener
{
	private static int	WAP = 3;
	
	private static final int HEIGHT	= WAP*WAP;
	private static final int WIDTH	= WAP*WAP;
	private static final int SIZE	= WAP;
	
	private static final int GEN_COMPLETE			= 1;
	private static final int GEN_CONTINUE			= 2;
	private static final int GEN_LIMIT_REACHED		= 3;
	
	private static final int GEN_HIT_LIMIT	= 100;
	
	private SudokuValueImpl[][]	values;
	private SudokuValueImpl[][]	reflected_values;
	
	private List		strategies = new ArrayList();
	
	private boolean value_change_in_progress	= false;
	private boolean value_change_outstanding	= false;
	
	private boolean	notifications_disabled		= false;
	
	private Random	random = new Random();
	
	private String		insoluble_detail;
	private List		auto_solve_values	= new ArrayList();
	private Object[][]	sod_hints;
	
	private List		listeners	= new ArrayList();
	
	public
	SudokuModelImpl()
	{
		values = new SudokuValueImpl[WIDTH][HEIGHT];
		
		reflected_values = new SudokuValueImpl[WIDTH][HEIGHT];
		
		for (int x=0;x<WIDTH;x++){
			
			for (int y=0;y<HEIGHT;y++){
				
				SudokuValueImpl	value = new SudokuValueImpl( this, x, y, SIZE*SIZE );
				
				value.addListener( this );
				
				values[x][y] = value;
				
				reflected_values[y][x] = value;
			}
		}

			// simple strategies; dice first as this is easier for users to comprehend when
			// generating hints than slice-based ones
		
		strategies.add( new SudokuStrategyDice( this ));
		strategies.add( new SudokuStrategySlice( this ));

			// complex strategies
		
		strategies.add( new SudokuStrategyTupleCycles( this ));
		strategies.add( new SudokuStrategySliceDiceRestriction( this ));
		
			// worst
		
		strategies.add( new SudokuStrategyXWing( this ));

	}
	
	public void
	load(
		int[][]	load_values )
	{
		try{
			notifications_disabled	= true;
				
			for (int x=0;x<WIDTH;x++){
				
				for (int y=0;y<HEIGHT;y++){
					
					SudokuValueImpl	value = values[x][y];
					
					value.clear();
					
					int	val = load_values[y][x];
					
					if ( val > 0 ){
						
						// System.out.println( "importing " + x + "/" + y );
						
						value.setValue( null, val, true, new SudokuValueMetricImpl( SudokuValueMetric.MT_CORE, "load" ) );
						
						value.setCore( true );
					}
				}
			}
		}finally{
			
			notifications_disabled	= false;	
		}
		
		solve();
	}

	protected int[][]
	exportValues()
	{
		int[][]	vals = new int[WIDTH][HEIGHT];
		
		for (int x=0;x<WIDTH;x++){
			
			for (int y=0;y<HEIGHT;y++){
	
				if ( values[x][y].isExplicit()){
					
					vals[y][x] = values[x][y].getValue();
				}
			}
		}
		
		return( vals );
	}
	
	protected void
	importValues(
		int[][]	values )
	{
		load( values );
	}
	
	public void
	exportValues(
		File		file )
	
		throws Exception
	{
		try{
			ObjectOutputStream	oos = new ObjectOutputStream( new FileOutputStream( file ));
			
			oos.writeInt( 1 );
			
			int[][]	vals = new int[WIDTH][HEIGHT];
			
			int[][][] overrides = new int[WIDTH][HEIGHT][];
			
			for (int x=0;x<WIDTH;x++){
				
				for (int y=0;y<HEIGHT;y++){
		
					SudokuValue	value = values[x][y];
					
					if (value.isExplicit()){
						
						vals[x][y] = value.isCore()?value.getValue():-value.getValue();
					}
					
					overrides[x][y] = value.getPossibleOverrides();
				}
			}
			
			oos.writeObject( vals );
			
			oos.writeObject( overrides );
			
			oos.close();
			
		}catch( Throwable e ){
			
			throw( new Exception( e.getMessage()));
		}
	}
	
	public void
	importValues(
		File		file )
	
		throws Exception
	{
		try{
			notifications_disabled = true;
			
			ObjectInputStream	ois = new ObjectInputStream( new FileInputStream( file ));

			int	version = ois.readInt();
			
			int[][]	vals = (int[][])ois.readObject();
			
			for (int x=0;x<WIDTH;x++){
				
				for (int y=0;y<HEIGHT;y++){
		
					SudokuValueImpl	value = values[x][y];

					int	v = vals[x][y];
					
					value.clear();

					if ( v != 0 ){
												
						if ( v > 0 ){
							
							value.setValue( null, v, true, new SudokuValueMetricImpl( SudokuValueMetric.MT_CORE, "load" ));
							
							value.setCore( true );
						}else{
							
							value.setValue( null, -v, true, new SudokuValueMetricImpl( SudokuValueMetric.MT_ENTERED, "load" ));
						}
					}
				}
			}
			
			if ( version >= 1 ){
				
				int[][][]	overrides = (int[][][])ois.readObject();

				for (int x=0;x<WIDTH;x++){
					
					for (int y=0;y<HEIGHT;y++){
			
						SudokuValueImpl	value = values[x][y];

						value.setPossibleOverrides( overrides[x][y] );
					}
				}
			}
			
			ois.close();
			
		}catch( Throwable e ){
			
			e.printStackTrace();
			
			throw( new Exception( e.getMessage()));
			
		}finally{
			
			notifications_disabled	= false;
			
			solve();
		}
	}
	
	public void
	solve()
	{
		solve( new SudokuSolveInstanceImpl( false, getHeight()*getWidth()));
	}
	
	protected void
	solve(
		SudokuSolveInstance solve_instance )
	{		
		solveSupport( solve_instance, null );
	}
	
	public void
	valueChanged(
		SudokuSolveInstance	solve_instance,
		SudokuValue			value )
	{
		solveSupport( solve_instance, value );
	}
	
	public void
	possibilityChanged(
		SudokuSolveInstance	solve_instance,
		SudokuValue			value,
		int					val,
		boolean				possible )
	{
	}
	
	public void
	otherChanged(
		SudokuValue			value )
	{
	}
	
	protected void
	solveSupport(
		SudokuSolveInstance		solve_instance,
		SudokuValue				value )
	{			
		if ( value_change_in_progress ){
			
				// ignore value resets

			if ( value != null && value.hasValue()){
			
				value_change_outstanding	= true;
				
				if ( !solve_instance.forGeneration()){
					
					Object[][]	hints = new Object[WIDTH][HEIGHT];
					
					for (int i=0;i<WIDTH;i++){
						for (int j=0;j<HEIGHT;j++){
							hints[i][j] = values[i][j].exportHints();
						}
					}
				
					auto_solve_values.add( new Object[]{ value, hints });
				}
			}
			
			return;
		}
		
		// System.out.println( "Solve starts" );
		
		auto_solve_values.clear();
		
		sod_hints	= null;
		
		try{
			value_change_in_progress	= true;
				
			resetSupport( false );
		
		}finally{
			
			value_change_in_progress	= false;
		}
		
		solveCycle( solve_instance );

		insoluble_detail = null;
		
		int	chain_depth = solve_instance.getChainDepth();
		
		if ( chain_depth > 0 ){
		
			String			prev_chain			= null;			

			int	cycle = 1;
			
			while ( chain_depth > 0 && 
					isValid() && 
					!( isSolved() || insoluble_detail != null  )){
				
				chain_depth--;
				
				// System.out.println( "cycle " + cycle++ );
				
				final SudokuValueImpl[][]	snapshot = new SudokuValueImpl[WIDTH][HEIGHT];
				
				for (int x=0;x<WIDTH;x++){					
					for (int y=0;y<HEIGHT;y++){
						snapshot[x][y] = values[x][y].exportAll();
					}
				}
				
				List	snapshot_solve_values = new ArrayList( auto_solve_values );
				
				final Set[]					all_change_set		= { null };
				final List[]				change_list 		= { null };
				final boolean[]				change_list_better	= { false };
				final int[]					change_list_best 	= { 9999999 };
				
				SudokuValueListener	listener = 
					new SudokuValueListener()
					{
						public void
						valueChanged(
							SudokuSolveInstance		solve_instance,
							SudokuValue				value )
						{
							all_change_set[0].add( value );
							
							if ( solve_instance.abandoned()){
								
								return;
							}
							
							List	changes = change_list[0];
							
							// System.out.println( "    val:" + value.getString(0));
						
							if ( changes.size() ==  change_list_best[0] - 1 ){
								
								solve_instance.abandon();
								
								return;
							}
							
							changes.add( value.getString(0)+"="+value.getValue());
							
							if ( !value.isValid()){
								
								change_list_best[0] = changes.size();
								
								change_list_better[0]	= true;
								
								solve_instance.abandon();	
							}
						}
						
						public void
						possibilityChanged(
							SudokuSolveInstance	solve_instance,
							SudokuValue			value,
							int					val,
							boolean				possible )
						{
							all_change_set[0].add( value );

							if ( solve_instance.abandoned()){
								
								return;
							}

							//System.out.println( "    poss:" + value.getString(0) + ":" + val );
														
							
							List	changes = change_list[0];

							if ( !value.isValid()){
								
								if ( changes.size() <  change_list_best[0] - 1 ){
											
									SudokuValue o_value = snapshot[value.getX()][value.getY()];
									
									boolean[]	possibles	= value.getPossibles(false);
									boolean[]	o_possibles = o_value.getPossibles(false);
									
									String	str = "";
										
									for (int i=0;i<possibles.length;i++){
										
										if ( o_possibles[i] && !possibles[i]){
											
											String reason = "";
											
											if ( value.getPossibleReasons()[i] != null ){
												
												reason = "[" + value.getPossibleReasons()[i].getReason() + "]";
											}
											
											str += "<BR>&nbsp;&nbsp;&nbsp;&nbsp;" + (i+1) + reason;
										}
									}
									
									// str += "</UL>";
									
									changes.add( value.getString(0) + ":" + str );

									change_list_best[0] = changes.size();
									
									change_list_better[0] = true;
								}
								
								solve_instance.abandon();
							}
						}
						public void
						otherChanged(
							SudokuValue			value )
						{
						}
					};
					
				for (int x=0;x<WIDTH;x++){					
					for (int y=0;y<HEIGHT;y++){
						values[x][y].addListener( listener );
					}
				}
					
				SudokuValueImpl	best_value			= null;
				int				best_possibility	= 0;
				String			best_reason			= "";
				
				try{
					for (int x=0;x<WIDTH && insoluble_detail == null ;x++){
						
						for (int y=0;y<HEIGHT &&  insoluble_detail == null ;y++){
	
							SudokuValueImpl	v = values[x][y];
							
							if ( !v.hasValue()){
								
								boolean[]	poss = (boolean[])v.getPossibles( false ).clone();
								
								int	num_ok	= -1;
								
								for (int p=0;p<poss.length;p++){
									
									if ( poss[p] ){
																				
										try{
											
											try{
												
												notifications_disabled = true;
											
												v.setValue( p+1, false );
											
											}finally{
												
												notifications_disabled	= false;
											}
											
											all_change_set[0]		= new HashSet();
											change_list[0] 			= new ArrayList();
											change_list_better[0]	= false;
											
											all_change_set[0].add( v );
											change_list[0].add( v.getString(0)+"="+v.getValue());
												
											auto_solve_values.clear();

											solveCycle( new SudokuSolveInstanceImpl( true, 0 ));
											
											if ( isSolved()){
												
												
												
												if ( num_ok != -1 ){
													
													insoluble_detail = "Insoluble: " + v.getString(0) + " can be either " + num_ok + " or " + (p+1);
																										
													break;
													
												}else{
													
												     num_ok = p+1;
												      
												}
											}
											
											// System.out.println( x + "," + y + ":" + p + " -> " + change_list_better[0] + ", change list = " + change_list[0].size());
											
											if ( change_list_better[0] ){
												
												List	c = change_list[0];
												
												String	str  = "";
																									
												for (int i=0;i<c.size();i++){
												
													str += (i==0?"":",") + c.get(i);
												}
												
												// System.out.println( "    "  + str );
												
												best_value 			= v;
												best_possibility	= p;
												best_reason			= str;
											}
																					
										}finally{
																							
											Iterator	it = all_change_set[0].iterator();
											
											while( it.hasNext()){
												
												SudokuValueImpl	q = (SudokuValueImpl)it.next();
											
												q.importAll(snapshot[q.getX()][q.getY()]);
												
											}
										}
									}
								}
							}
						}
					}
				}finally{
					
					for (int x=0;x<WIDTH;x++){					
						for (int y=0;y<HEIGHT;y++){
							values[x][y].removeListener( listener );
						}
					}
				}
				
				auto_solve_values = snapshot_solve_values;
				
				if ( best_value != null ){
					
					try{
						
						notifications_disabled = true;
					
						best_value.setPossible( 
								null,
								best_possibility+1, false,
								new SudokuPossibilityMetricImpl(
										SudokuPossibilityMetric.MT_CHAIN,
										(prev_chain==null?"":("[see " + prev_chain + "] ")) + best_reason ));
				
						prev_chain	= best_value.getString(0);
						
					}finally{
						
						notifications_disabled	= false;
					}								
					
					solveCycle( solve_instance );
					
				}else{
					
					break;
				}
			}
		}
		
		if ( !solve_instance.forGeneration()){
			
			setHintState();
		
			fireModelChanged();
		}
		
		// System.out.println( "Solve complete: chain = " + auto_solve_values.size());
	}
	
	protected void
	solveCycle(
		SudokuSolveInstance	solve_instance ) 
	{		
		while( true && !solve_instance.abandoned()){
			
			try{
				value_change_in_progress	= true;
															
					// apply simple ones first
				
				for ( int comp=0;comp<SudokuStrategy.COMPLEXITIES.length && !solve_instance.abandoned();comp++){
					
					int	type = SudokuStrategy.COMPLEXITIES[comp];
					
					for (int i=0;i<strategies.size() && !solve_instance.abandoned();i++){
						
						SudokuStrategy	strategy = (SudokuStrategy)strategies.get(i);
						
						if ( strategy.getComplexity() == type ){
							
							strategy.apply( solve_instance );
						}
					}
					
					if ( sod_hints == null ){
					
						Object[][]	hints = new Object[WIDTH][HEIGHT];
						
						for (int i=0;i<WIDTH;i++){
							for (int j=0;j<HEIGHT;j++){
								hints[i][j] = values[i][j].exportHints();
							}
						}
						
						sod_hints = hints;
					}
					
					for (int i=0;i<strategies.size() && !solve_instance.abandoned();i++){
						
						SudokuStrategy	strategy = (SudokuStrategy)strategies.get(i);
																						
						strategy.update( solve_instance );
					}
					
					if ( value_change_outstanding ){
						
						break;
					}
				}
				
			}finally{
				
				value_change_in_progress	= false;
				
				if ( value_change_outstanding ){
					
					value_change_outstanding	= false;
					
				}else{
					
					break;
				}
			}
		}
	}
	
	public int
	getWidth()
	{
		return( WIDTH );
	}
	
	public int
	getHeight()
	{
		return( HEIGHT );
	}
	
	public int
	getSize()
	{
		return( SIZE );
	}
	
	public SudokuValue[][]
	getValues()
	{
		return( values );
	}
	
	public SudokuValue[][]
 	getReflectedValues()
 	{
 		return( reflected_values );
 	}

	protected boolean
	disableNotifications()
	{
		return( notifications_disabled );
	}
	
	public boolean
	isSolved()
	{
		for (int x=0;x<WIDTH;x++){
			
			for (int y=0;y<HEIGHT;y++){
				
				if ( !values[x][y].isValid()){
					
					return( false );
				}
				
				if ( !values[x][y].hasValue()){
					
					return( false );
				}
			}
		}
		
		return( true );
	}
	
	
	public boolean
	isValid()
	{
		for (int x=0;x<WIDTH;x++){
			
			for (int y=0;y<HEIGHT;y++){
				
				if ( !values[x][y].isValid()){
					
					return( false );
				}
			}
		}
		
		return( true );
	}
	
	public void
	getHint()
	{
		for (int h=0;h<auto_solve_values.size();h++){
				
			Object[]	x = (Object[])auto_solve_values.get(h);
				
			SudokuValueImpl	value = (SudokuValueImpl)x[0];
			
			if ( !value.isHint()){
					
				value.setHint(true);
					
				setHintState();
				
				return;
			}
		}
	}
	
	public void
	removeHint(
		SudokuValueImpl	value )
	{
		for (int h=0;h<auto_solve_values.size();h++){

			Object[]	x = (Object[])auto_solve_values.get(h);
		
			SudokuValueImpl	v = (SudokuValueImpl)x[0];
		
			if ( v == value ){
						
				for (int j=h;j<auto_solve_values.size();j++){
					
					x = (Object[])auto_solve_values.get(j);
					
					v = (SudokuValueImpl)x[0];

					if ( v.isHint()){
						
						v.setHint( false );
					}
				}
				
				setHintState();
				
				break;
			}
		}
	}
	
	protected void
	setHintState()
	{
		Object[][]	prev_hints	= null;
			
		for (int h=0;h<auto_solve_values.size();h++){
						
			Object[]	x = (Object[])auto_solve_values.get(h);
						
			SudokuValueImpl	value = (SudokuValueImpl)x[0];
					
			if ( value.isHint()){
						
				prev_hints	= (Object[][])x[1];
					
			}else{
	
				if ( prev_hints != null ){
					
					for (int i=0;i<WIDTH;i++){
						for (int j=0;j<HEIGHT;j++){
							values[i][j].importHints( prev_hints[i][j]);
						}
					}
					
					return;
				}
			}
		}
		
		if ( sod_hints != null ){
			
			for (int i=0;i<WIDTH;i++){
				for (int j=0;j<HEIGHT;j++){
					values[i][j].importHints( sod_hints[i][j]);
				}
			}
		}
	}
	
	public void
	reset()
	{
		resetSupport( true );
		
		solve();
	}
	
	protected void
	resetSupport(
		boolean reset_non_core )
	{
		boolean	set = false;
		
		if ( !value_change_in_progress ){
		
			set = true;
			
			value_change_in_progress = true;
		}
		
		try{
			for (int x=0;x<WIDTH;x++){
				
				for (int y=0;y<HEIGHT;y++){
										
					values[x][y].reset( reset_non_core );
				}
			}
		}finally{
			
			if ( set ){
				
				value_change_in_progress	= false;
				value_change_outstanding	= false;
				
			}
		}
	}
	
	public void
	clear()
	{
		try{
			notifications_disabled	= true;

			for (int x=0;x<WIDTH;x++){
				
				for (int y=0;y<HEIGHT;y++){
										
					values[x][y].clear();
				}
			}
		}finally{
			
			notifications_disabled	= false;
		}
		
		solve();
	}
	
	public void
	generate(
		boolean								use_chains,
		int									symmetry,
		SudokuModelGenerationListener		listener )
	{
		int[]		best_complexity 	= {0};
		int[][][]	best_values 		= {exportValues()};
			
		int	cycle = 0;
		
		while( true ){
			
			cycle++;
			
			// System.out.println( "Generation cycle " + cycle );
			
			SudokuSolveInstance	solve_instance = new SudokuSolveInstanceImpl( true, use_chains?8:0 );
			
			try{
				notifications_disabled	= true;
				
				while( true ){
					
					if ( generateSupportOld( solve_instance )){
						
						//print();
						
						break;
					}
				}
				
			}finally{
				
				notifications_disabled = false;
			}
			
			List	remaining = new ArrayList();
			
			for (int x=0;x<WIDTH;x++){
				
				for( int y=0;y<HEIGHT;y++){
	
					remaining.add( values[x][y] );
				}
			}
			
			int[]	gen_count = {0};
			
			int	res = prune( solve_instance, use_chains, symmetry, listener, new HashSet(), 0, remaining, best_complexity, best_values, gen_count );
			
			if ( res == GEN_COMPLETE ){
				
				break;
			}
		}
		
		importValues( best_values[0]);
	}
	
	public SudokuModelComplexity
	getComplexity()
	{
		final int		value_count;
		final int		complex_rule_count;
		final int		complexity_value;
		final String	complexity_desc;
		final String	complexity_detail;
		
		if ( isSolved() && insoluble_detail == null ){
			
			Map	metric_map = new HashMap();
			
			int	num_count = 0;
			
			int	complex	= 0;
			
			for (int x=0;x<WIDTH;x++){
				
				for( int y=0;y<HEIGHT;y++){
	
					SudokuValueImpl	value = values[x][y];
					
					if ( value.hasValue() && value.isExplicit()){
						
						num_count++;
					}
					
					SudokuPossibilityMetric[]	metrics = value.getPossibleReasons();
					
					for (int i=0;i<metrics.length;i++){
						
						SudokuPossibilityMetric	metric = metrics[i];
						
						if ( metric != null ){
							
							int		type = metric.getType();
							
							if ( 	type == SudokuPossibilityMetric.MT_ALREADY_IN_CHUNK ||
									type == SudokuPossibilityMetric.MT_ALREADY_IN_SLICE ){
								
									// boring
								
							}else if ( type == SudokuPossibilityMetric.MT_X_WING ){
								
								complex += 25;
								
							}else if ( type == SudokuPossibilityMetric.MT_CHAIN ){
								
								complex += 100;
								
							}else{
								
								complex++;
							}
							
							Integer	count = (Integer)metric_map.get(new Integer( type));
						
							if ( count == null ){
								
								count = new Integer(1);
								
							}else{
								
								count = new Integer( count.intValue()+1 );
							}
							
							metric_map.put( new Integer(type), count);
						}
					}
				}
			}
			
			Iterator	it = metric_map.keySet().iterator();
			
			String	str = "Values = " + num_count +"\n";
			
			while( it.hasNext()){
				
				Integer	type = (Integer)it.next();
				
				Integer count = (Integer)metric_map.get(type);
				
				str += "  " + SudokuPossibilityMetric.MT_NAMES[type.intValue()] + " -> " + count + "\n";
			}
			
			value_count			= num_count;
			complex_rule_count	= complex;
			complexity_value	= 81 - num_count + (10*complex); 
			complexity_desc		= "Values = " + num_count + ", complexity = " + complex;
			complexity_detail	= str;
		}else{
			
			value_count				= 0;
			complex_rule_count		= 0;
			complexity_value		= 0;
			complexity_desc			= insoluble_detail==null?"Insoluble":insoluble_detail;
			complexity_detail		= complexity_desc;
		}
		
		return( 
			new SudokuModelComplexity()
			{
				public int
				getValues()
				{
					return( value_count );
				}
				
				public int
				getComplexRuleCount()
				{
					return( complex_rule_count );
				}
				
				public int
				getComplexityMetric()
				{
					return( complexity_value );
				}
				
				public String
				getString(
					boolean	detailed )
				{
					return( detailed?complexity_detail:complexity_desc );
				}
			});
	}
	
	protected int
	prune(
		SudokuSolveInstance				solve_instance,
		boolean							use_chains,
		int								symmetry,
		SudokuModelGenerationListener	listener,
		Set								fingerprints,
		int								depth,
		List							remaining,
		int[]							best_complexity,
		int[][][]						best_values,
		int[]							gen_count )
	{
		if ( depth >= 40 ){
			
			StringBuffer	fingerprint = new StringBuffer();
			
			for (int i=0;i<remaining.size();i++){
				
				SudokuValue	value = (SudokuValue)remaining.get(i);				

				fingerprint.append((char)(value.getX()*WIDTH+value.getY()));
			}
			
			String	fps = fingerprint.toString();
			
			// System.out.println( fingerprint );
			
			if ( fingerprints.contains(fps)){
							
				return( GEN_CONTINUE );
			}
			
			fingerprints.add( fps );
		}
		
		List<SudokuValueImpl>	copy = new ArrayList<SudokuValueImpl>(remaining);
		
		// System.out.println( "depth=" + depth + ",rem=" + remaining.size());
		
		int	limit;
		
		if ( depth < 10 ){
			
			limit = 2;
				
		}else if ( depth < 50 ){
			
			limit = 2;
			
		}else{
			
			limit = copy.size();
		}
		
		for (int i=0;copy.size()>0&& i<limit;i++){
			
			List<SudokuValueImpl> to_removes = new ArrayList<SudokuValueImpl>();
			
			Map<SudokuValueImpl,int[]> trm_details = new HashMap<SudokuValueImpl, int[]>();
			
			SudokuValueImpl	to_remove_seed = (SudokuValueImpl)copy.remove( random.nextInt( copy.size()));
				
			to_removes.add( to_remove_seed );

			List<int[]> others_to_remove = new ArrayList<int[]>();
						
			if ( symmetry == SYM_HORIZONTAL ){
				
				int x = to_remove_seed.getX();
	
				int	other_x = WIDTH - x - 1;
				
				if ( other_x != x ){
	
					others_to_remove.add( new int[]{ other_x, to_remove_seed.getY()} );
				}
			}else if ( symmetry == SYM_VERTICAL ){
				
				int y = to_remove_seed.getY();
	
				int	other_y = HEIGHT - y - 1;
				
				if ( other_y != y ){
	
					others_to_remove.add( new int[]{ to_remove_seed.getX(), other_y } );
				}
			}else if ( symmetry == SYM_SQUARE ){
				
				int y = to_remove_seed.getY();
	
				int	other_y = HEIGHT - y - 1;
				
				int x = to_remove_seed.getX();
				
				int	other_x = WIDTH - x - 1;

				if ( other_y != y ){
					
					others_to_remove.add( new int[]{ to_remove_seed.getX(), other_y } );
				}
				
				if ( other_x != x ){
	
					others_to_remove.add( new int[]{ other_x, to_remove_seed.getY() } );
				}
				
				if ( other_y != y &&  other_x != x ){
					
					others_to_remove.add( new int[]{ other_x, other_y } );
				}
			}
			
			for ( int[] coords: others_to_remove ){
				
				boolean found = false;
				
				for ( SudokuValueImpl temp: copy ){
					
					if ( temp.getX() == coords[0] && temp.getY() == coords[1]){
						
						to_removes.add( temp );
						
						copy.remove( temp );

						found = true;
						
						break;
					}
				}
					
				if ( !found ){
					
					System.out.println( "eh: " +  to_remove_seed.getY() + ": " + coords[0]  + "/" +  coords[1] );
				}
			}
			

			
			for ( SudokuValueImpl to_remove: to_removes ){
			
				int	index = remaining.indexOf( to_remove );
			
				remaining.remove( index );
			
				int	old = to_remove.getValue();
				
				trm_details.put( to_remove, new int[]{ index, old });
			}
			
			// System.out.println( "removing: " + to_remove.getString(2));

			try{				
				try{
					notifications_disabled	= true;
						
					for ( SudokuValueImpl to_remove: to_removes ){
						
						to_remove.clear();
					}
				}finally{
						
					notifications_disabled	= false;
				}
							
					// only kick in possible chains if we've gone a fair way...
				
				int	chain_depth = solve_instance.getChainDepth();
				
				if ( chain_depth > 0 ){
					
					int	target_depth = 0;
					
					if ( depth > 50 && random.nextInt(5) == 0 ){
						
						target_depth = chain_depth;
					}
					
					solve_instance.setChainDepth( target_depth );
				}
				
				solve( solve_instance );
					
				solve_instance.setChainDepth( chain_depth );
				
				if ( isSolved()){
					
					int	complexity = getComplexity().getComplexityMetric();
					
					if ( complexity > best_complexity[0]){
												
						best_complexity[0] = complexity;
						
						best_values[0] = exportValues();
												
						// System.out.println( "   best = " + depth );
					}
					
					if ( !listener.generated()){
						
						return( GEN_COMPLETE );
					}

					gen_count[0]++;
					
					if ( gen_count[0] == GEN_HIT_LIMIT ){
						
						return( GEN_LIMIT_REACHED );
					}
					
					int res = prune( solve_instance, use_chains, symmetry, listener, fingerprints, depth+1, remaining, best_complexity, best_values, gen_count );
					
					if ( res != GEN_CONTINUE ){
						
						return( res );
					}
				}else{
					
					if ( !listener.generated()){
						
						return( GEN_COMPLETE );
					}		
				}
				
			}finally{
			
				try{
					notifications_disabled	= true;
						
					for ( SudokuValueImpl to_remove: to_removes ){
						
						to_remove.setValue( solve_instance, trm_details.get( to_remove )[1], true, new SudokuValueMetricImpl( SudokuValueMetric.MT_CORE, "generate" ));
					}
					
				}finally{
						
					notifications_disabled	= false;
				}
				
				Collections.reverse( to_removes );
				
				for ( SudokuValueImpl to_remove: to_removes ){
				
					remaining.add( trm_details.get( to_remove )[0], to_remove );
				}
			}
		}
		
		return( GEN_CONTINUE );
	}
	

	protected boolean
	generateSupportNew(
		SudokuSolveInstance	solve_instance )
	{
			// randomly fill the model with valid data
				
		for (int x=0;x<WIDTH;x++){
			
			for (int y=0;y<HEIGHT;y++){
									
				values[x][y].setValue( solve_instance, -1, true, new SudokuValueMetricImpl( SudokuValueMetric.MT_CORE, "generate" ));
				
				values[x][y].setAllPossible();
			}
		}
		
		return( xxx( solve_instance, 0 ));
	}
	
	protected boolean
	xxx(
		SudokuSolveInstance	solve_instance,
		int					pos )
	{
		if ( pos == WIDTH*HEIGHT ){
						
			return( true );
		}
				
		int	x = pos/WIDTH;
		int	y = pos%HEIGHT;
		
		SudokuValueImpl	value = values[x][y];
		
		boolean	[]	possibles = (boolean[])value.getPossibles(false).clone();
		
		List	l = new ArrayList();
		
		for (int i=0;i<possibles.length;i++){
			
			if ( possibles[i] ){
				
				l.add( new Integer(i));
				
			}
		}
		
		while( l.size() > 0 ){
		
			int	i = ((Integer)l.remove( random.nextInt(l.size()))).intValue();
				
			value.setValue( 
				solve_instance,
				i+1, 
				true, 
				new SudokuValueMetricImpl( SudokuValueMetric.MT_CORE, "generate" ));
			
			List	changed = new ArrayList();
			
			int	x_chunk = (x/SIZE)*SIZE;
			int	y_chunk = (y/SIZE)*SIZE;
			
			for (int xc=x_chunk;xc<x_chunk+SIZE;xc++){
						
				for (int yc=y_chunk;yc<y_chunk+SIZE;yc++){
					
					SudokuValueImpl	v = (SudokuValueImpl)values[xc][yc];

					if ( v.isPossible(i+1)){
						v.setPossible( solve_instance, i+1, false, null );
						changed.add(v);
					}
				}
			}	
			
			for (int z=0;z<WIDTH;z++){
									
				SudokuValueImpl	v = (SudokuValueImpl)values[z][y];

				if ( v.isPossible(i+1)){
					v.setPossible( solve_instance, i+1, false, null );
					changed.add(v);
				}
				
				v = (SudokuValueImpl)values[x][z];

				if ( v.isPossible(i+1)){
					v.setPossible( solve_instance, i+1, false, null );
					changed.add(v);
				}
			}
			
			if ( xxx( solve_instance, pos+1 )){
				
				return( true );
			}
			
			for (int j=0;j<changed.size();j++){
				((SudokuValueImpl)changed.get(j)).setPossible( solve_instance, i+1, true, null );
			}
		}
		
		value.setValue( -1, false );
		
		return( false );
	}
	
	protected boolean
	generateSupportOld(
		SudokuSolveInstance	solve_instance )
	{
			// randomly fill the model with valid data
				
		for (int x=0;x<WIDTH;x++){
			
			for (int y=0;y<HEIGHT;y++){
									
				values[x][y].setValue( solve_instance, -1, true, new SudokuValueMetricImpl( SudokuValueMetric.MT_CORE, "generate" ));
			}
		}
		
		for (int i=0;i<SIZE*SIZE;i++){
			
				// place each number randomly in a dice
			
			for (int x=0;x<WIDTH;x+=SIZE){
				
				for (int y=0;y<HEIGHT;y+=SIZE){
		
					boolean	valid[] = new boolean[SIZE*SIZE];
					
					Arrays.fill( valid, true );
	
					for (int x_offset=0;x_offset<SIZE;x_offset++){
						
						for (int y_offset=0;y_offset<SIZE;y_offset++){
							
							if ( values[x+x_offset][y+y_offset].hasValue()){
								
								valid[x_offset*SIZE+y_offset] = false;
							}
						
							for (int r=0;r<SIZE*SIZE;r++){
								
								if ( values[x+x_offset][r].getValue() == i+1 ){
									
									valid[x_offset*SIZE+y_offset] = false;
									break;
								}
								if ( values[r][y+y_offset].getValue() == i+1 ){
									
									valid[x_offset*SIZE+y_offset] = false;
									break;
								}							
							}
						}
					}
						
					int	ok = 0;
						
					for (int p=0;p<valid.length;p++){
							
						if ( valid[p] ){
								
							ok++;
						}
					}
						
					if ( ok == 0 ){
													
						return( false );
					}
					
					int	ran = random.nextInt(ok);
						
					int	selected = 0;
						
					for (int p=0;p<valid.length;p++){
							
						if ( valid[p]){
								
							if ( ran == 0 ){
									
								selected = p;
									
								break;
							}
								
							ran--;
						}
					}
					
					values[x+(selected/SIZE)][y+(selected%SIZE)].setValue(
							solve_instance,
							i+1, 
							true, 
							new SudokuValueMetricImpl( SudokuValueMetric.MT_CORE, "generate" ));
				}
			}
		}
		
		return( true );
	}
	protected void
	removePossibilities(
		SudokuSolveInstance	solve_instance,
		int					x,
		int					y,
		int					v )
	{
		int	x_chunk = (x/SIZE)*SIZE;
		int	y_chunk = (y/SIZE)*SIZE;
		
		for (int xc=x_chunk;xc<x_chunk+SIZE;xc++){
					
			for (int yc=y_chunk;yc<y_chunk+SIZE;yc++){
						
				if ( xc != x && yc != y ){
				
					SudokuValueImpl	value = (SudokuValueImpl)values[xc][yc];
					
					if ( value.isPossible(v)){
						
						value.setPossible( 
							solve_instance,
							v, false, 
							new SudokuPossibilityMetricImpl(
									SudokuPossibilityMetric.MT_ALREADY_IN_CHUNK,
									"already in chunk" ));
					}
				}
			}
		}
				
		
		for (int z=0;z<WIDTH;z++){
				
			if ( z != x ){
				
				SudokuValueImpl	value = (SudokuValueImpl)values[z][y];

				if ( value.isPossible(v)){

					value.setPossible( 
							solve_instance,
							v, false,
							new SudokuPossibilityMetricImpl(
									SudokuPossibilityMetric.MT_ALREADY_IN_SLICE,
									"already in row" ));
				}
			}
			if ( z != y ){
				
				SudokuValueImpl	value = (SudokuValueImpl)values[x][z];

				if ( value.isPossible(v)){
					
					value.setPossible( 
						solve_instance,
						v, false, 
						new SudokuPossibilityMetricImpl(
								SudokuPossibilityMetric.MT_ALREADY_IN_SLICE,
								"already in column" ));
				}
			}
		}
	}
	
	protected void
	fireModelChanged()
	{
		for (int i=0;i<listeners.size();i++){
			
			((SudokuModelListener)listeners.get(i)).modelChanged();
		}
	}
	
	public void
	addListener(
		SudokuModelListener	l )
	{
		listeners.add(l);
	}
	
	public void
	removeListener(
		SudokuModelListener	l )
	{
		listeners.remove(l);
	}
	
	protected void
	print()
	{
		System.out.println( "Print" );
		
		for (int x=0;x<WIDTH;x++){
			
			String	str = "    ";
			
			for (int y=0;y<HEIGHT;y++){
									
				int	val = values[x][y].getValue();
				
				String	s = val==-1?" ":Integer.toHexString( val );
				
				str += (y==0?"":" ") + s;
			}
			
			System.out.println( str );
		}
	}
}
